/*******************************************************************************
 *  bloggie.c   |   version 1.1     |   2-Clause BSD License    |   2014-03-09
 *  James Hendrie                   |   hendrie dot james at gmail dot com
 *
 *  Description:    Reads text from a file or stdin, prints HTML to a file or
 *                  stdout.  By default, it only replaces newline characters
 *                  with '<br/>\n', but other options allow the user to create
 *                  a functional HTML document from a text file.
*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#define LONG_STRING 256
#define SHORT_STRING 10


/*  Optstring
 *      h   help
 *      i   input file
 *      o   output file
 *      H   create an entire boring HTML document from text
 *      v   HTML version:  0 (default, none specified), 4 or 5
 *      s   font size, in 'em'
 *      S   header size for title, if displayed in page
 *      f   font family
 *      t   page title, if any
 *      T   display title at top of page
 *      B   background color
 *      F   forground color
 *      p   use 'pre' tag, good for code
 *      c   specify CSS file
 *      j   specify javascript file
 *      l   specify HTML language (en by default)
 *      -   read from stdin
 */
static const char *optString = "hH-Tpi:o:v:s:S:f:t:B:F:c:j:l:";


/*  Global options */
int htmlCreate;
int htmlVersion;
float fontSize;
int headerSize;
int displayTitle;
int usePre;

/*  Character pointers, to be filled later if necessary */
char *fontFamily;
char *pageTitle;
char *bgColor;
char *fgColor;
char *cssFile;
char *jsFile;
char *lang;


/*
 *  ============================    PRINT HELP  ================================
 */
void print_help(void)
{
    printf("Usage:  bloggie [OPTIONS]\n\n");
    printf("Options:\n");
    printf("  -h or --help\t\tPrint this help text\n");
    printf("  --version\t\tPrint version and author info\n");
    printf("  -i FILE\t\tSpecify input file (default stdin)\n");
    printf("  -o FILE\t\tSpecify output file (default stdout)\n");
    printf("  -\t\t\tRead input from stdin (same as no args)\n");
    printf("  -H\t\t\tWrite an HTML document, allows for style elements\n");
    printf("  -v\t\t\tHTML version (none, 3, 4 or 5).  Changes HTML header\n");
    printf("  -B\t\t\tBackground color; FFFFFF by default\n");
    printf("  -F\t\t\tForeground (text) color; 000000 by default\n");
    printf("  -f\t\t\tFont family\n");
    printf("  -s\t\t\tFont size; 'em' units for HTML version 4 or 5\n");
    printf("  -t\t\t\tPage title, if any\n");
    printf("  -T\t\t\tDisplay page title at top of page as header, if any\n");
    printf("  -S\t\t\tHeader size, if applicable.  3 = <h3>, 4 = <h4>, etc.\n");
    printf("  -p\t\t\tUse <pre> tag; good for code\n");
    printf("  -c\t\t\tSpecify CSS file (html versions 4 or 5)\n");
    printf("  -j\t\t\tSpecify Javascript file (html versions 4 or 5)\n");

    printf("\nThis program will read text, from either stdin or a file, and ");
    printf("print that text to\nanother file, stdout by default.  It can also");
    printf(" do other stuff.");
    printf("\nI use this to format HTML when read into Vim.  You can use it ");
    printf("for whatever.\n");
}



/*
 *  ============================    PRINT VERSION  =============================
 */
void print_version(void)
{
    printf("bloggie 1.1\n");
    printf("James Hendrie (hendrie dot james at gmail dot com)\n");
}



/*
 *  ============================    MEMORY ERROR  ==============================
 */
void mem_error( const char *desc )
{
    fprintf(stderr, "Error:  Out of memory\n");
    fprintf(stderr, "%s\n", desc);
}



/*
 *  ============================    STRING DEFAULT  ============================
 */
void string_default( char *sp, const char *string, size_t sizeDefault )
{
    if( sp[0] == '\0' )
        strncpy( sp, string, ( sizeDefault - strlen(string) - 1));
}



/*
 *  ============================    HTML BEGIN  ================================
 */
int html_begin( FILE *fp )
{
    /*  Init the string pointers, again */
    string_default( lang, "en", (SHORT_STRING));
    string_default( bgColor, "FFFFFF", (SHORT_STRING));
    string_default( fgColor, "000000", (SHORT_STRING));


    /*
     *  ---------------------   HEADER STUFF    -----------------------------
     */
    if( htmlVersion == 3 )
    {
        fputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n",fp);

        fputs("    <head>\n", fp);

        if( pageTitle != NULL )
            fprintf( fp, "    <title>%s</title>\n", pageTitle);

        fputs("</head>\n\n", fp);


    }
    else if( htmlVersion == 4 )
    {
        fputs("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n", fp);
        fputs("    \"http://www.w3.org/TR/html4/strict.dtd\">\n", fp);

        /*  Insert language, english if unspecified */
        fprintf( fp, "<html lang=\"%s\">\n", lang);

        fputs("  <head>\n", fp);
        fputs("    <meta http-equiv=\"content-type\" content=", fp);
        fputs("\"text/html; charset=utf-8\">\n", fp);

        /*  Title insertion */
        if( pageTitle != NULL )
            fprintf( fp, "    <title>%s</title>\n", pageTitle);

        /*  CSS insertion */
        if( cssFile != NULL )
        {
            fputs("    <link rel=\"stylesheet\" type=\"text/css\" ", fp);
            fprintf( fp, "href=\"%s\">\n", cssFile);
        }

        /*  Javascript insertion */
        if( jsFile != NULL )
        {
            fputs("    <script type=\"text/javascript\" ", fp);
            fprintf( fp, "src=\"%s\"></script>\n", jsFile);
        }

    }
    else if( htmlVersion == 5 )
    {
        fputs("<!DOCTYPE html>\n", fp);

        /*  Language insertion */
        fprintf( fp, "<html lang=\"%s\">\n", lang);

        fputs("  <head>\n", fp);
        fputs("    <meta charset=\"utf-8\">", fp);

        /*  Title insertion */
        if( pageTitle != NULL )
            fprintf( fp, "    <title>%s</title>\n", pageTitle);

        /*  CSS insertion */
        if( cssFile != NULL )
            fprintf(fp, "    <link rel=\"stylesheet\" href=\"%s\">\n", cssFile);

        /*  Javascript insertion */
        if( jsFile != NULL )
            fprintf( fp, "    <script src=\"%s\"></script>\n", jsFile);
    }
    else
    {
        fputs("<html>\n", fp);
        fputs("<head>\n", fp);
        if( pageTitle != NULL )
            fprintf( fp, "<title>%s</title>\n", pageTitle);
        fputs("</head>\n\n", fp);
    }


    /*
     * ------------------   STYLE   --------------------------
     */
    if( htmlVersion == 4 || htmlVersion == 5 )  //  html 4 or 5
    {
        fputs("<body style=\"", fp);
        if( bgColor != NULL )
            fprintf(fp, "background-color:%s;", bgColor);
        if( fgColor != NULL )
            fprintf(fp, "color:%s;", fgColor );
        if( fontFamily != NULL )
            fprintf(fp, "font-family:%s;", fontFamily);
        if( fontSize != 0 )
            fprintf(fp, "font-size:%f;", fontSize);
        fputs("\">\n\n", fp);

        /*  Finally, start the 'paragraph' */
        fputs("<p>\n", fp);

    }
    else    //  html 3.2 or 'other'
    {
        fprintf( fp, "<body bgcolor=\"#%s\">\n\n", bgColor);

        /*  Font stuff */
        if( fontSize != 0 || fgColor != NULL || fontFamily != NULL )
        {
            fputs("<font", fp);
            if( fontSize != 0 )
                fprintf( fp, " size=\"%d\"", (int)fontSize);
            if( fgColor != NULL )
                fprintf(fp, " color=\"#%s\"", fgColor );
            if( fontFamily != NULL )
                fprintf(fp, " face=\"%s\"", fontFamily );
            fputs(">\n\n", fp); //  Remember to close this font tag later
        }

    }


    /*  Insert header into page, if appropriate */
    if( pageTitle != NULL && displayTitle == 1 )
    {
        fprintf( fp, "<h%d>%s</h%d>\n", headerSize, pageTitle, headerSize);
    }

    /*  If they wanted to preserver formatting */
    if( usePre == 1 )
        fputs("<pre>\n", fp);


    return(0);
}



/*
 *  ============================    HTML END  ==================================
 */
int html_end( FILE *fp )
{
    if( usePre == 1 )
        fputs("</pre>\n", fp);

    if( htmlVersion == 4 || htmlVersion == 5 )
        fputs("\n</p>\n</body>\n</html>\n", fp);
    else
        fputs("\n</font>\n</body>\n</html>\n", fp);

    return(0);
}



/*
 *  ============================    BLOGGIE  ===================================
 */
int bloggie( FILE *inFile, FILE *outFile )
{
    /*  Variable to hold our character */
    int ch;

    /*  While there's something to read from the file */
    while( ( ch = fgetc( inFile ) ) != EOF )
    {
        /*
         * If we hit a newline character, put a string of text that includes the
         * <br/> statement into the file.  Otherwise, just put the character
         * that was read into it.
         */
        if( ch == '\n' )
            fputs("<br/>\n", outFile);
        else
            fputc( ch, outFile );
    }

    return(0);
}



/*
 *  ============================    CHECK STRING  ==============================
 */
void check_string( char *sp )
{
    if( sp == NULL )
    {
        mem_error("in check_string()");
        exit(1);
    }
}



/*
 * -----------------------------------------------------------------------------
 *  ============================    MAIN  ======================================
 * -----------------------------------------------------------------------------
 */
int main( int argc, char *argv[] )
{
    /*  Create our file pointers */
    FILE *inFile = NULL;
    FILE *outFile = NULL;

    /*  Initialize our global integers */
    htmlCreate = 0;
    htmlVersion = 0;
    fontSize = 0;
    headerSize = 3;
    displayTitle = 0;
    usePre = 0;

    /*  Same with global string pointers */
    lang = malloc( (SHORT_STRING) );
    check_string( lang );

    bgColor = malloc( (SHORT_STRING) );
    check_string( bgColor );

    fgColor = malloc( (SHORT_STRING) );
    check_string( fgColor );

    pageTitle = malloc( (LONG_STRING) );
    check_string( pageTitle );

    fontFamily = malloc( (LONG_STRING) );
    check_string( fontFamily );

    cssFile = malloc( (LONG_STRING) );
    check_string( cssFile );

    jsFile = malloc( (LONG_STRING) );
    check_string( jsFile );

    /*  ----------------    ARGUMENT CHECKING   ------------------ */

    /*  Long opts */
    if( argc > 1 )
    {
        if( strcmp( argv[1], "--help") == 0 )
        {
            print_help();
            return(0);
        }
        else if( strcmp( argv[1], "--version") == 0 )
        {
            print_version();
            return(0);
        }
    }


    /*  Short opts */
    int opt = 0;
    opt = getopt( argc, argv, optString );
    while( opt != -1 )
    {
        /*  Check and see if we got a wise-aleck */
        if( ! isalpha(opt) )
        {
            return(1);
        }

        switch( opt )
        {
            case 'h':   //  Print help
                print_help();
                return(0);
                break;
            case 'i':   //  Input file specified
                inFile = fopen( optarg, "r" );
                break;
            case '-':   //  Read from stdin
                inFile = stdin;
                break;
            case 'o':   //  Output file specified
                outFile = fopen( optarg, "w+" );
                break;
            case 'H':   //  Create html
                htmlCreate = 1;
                break;
            case 'v':   //  html version
                htmlVersion = atoi( optarg );
                break;
            case 's':   //  font size in em
                fontSize = atof( optarg );
                break;
            case 'S':   //  Size of header tags (3 for <h3>, 4 for <h4>, etc.)
                headerSize = atoi( optarg );
                break;
            case 'f':   //  font family
                strncpy( fontFamily, optarg, (LONG_STRING));
                break;
            case 't':   //  page title
                strncpy( pageTitle, optarg, (LONG_STRING));
                break;
            case 'T':   //  display title at top of page in header tags
                displayTitle = 1;
                break;
            case 'B':   //  Background color of page
                strncpy( bgColor, optarg, (SHORT_STRING));
                break;
            case 'F':   //  Foreground (text) color
                strncpy( fgColor, optarg, (SHORT_STRING));
                break;
            case 'p':   //  wrap all text in 'pre' HTML tag
                usePre = 1;
                break;
            case 'c':   //  user specifies a CSS file
                strncpy( cssFile, optarg, (LONG_STRING));
                break;
            case 'j':   //  user specifies a javascript file
                strncpy( jsFile, optarg, (LONG_STRING));
                break;
            case 'l':
                strncpy( lang, optarg, (SHORT_STRING));
            default:
                /*  Won't be seeing this, disregard */
                break;
        }

        opt = getopt( argc, argv, optString );
    }


    /*  Double check the file pointer.  If it's null, try to point to stdin */
    if( inFile == NULL )
    {
        inFile = stdin;
        if( inFile == NULL )
        {
            fprintf(stderr, "ERROR:  Cannot open stdin for reading\n");
            return(1);
        }
    }

    /*  Same for stdout */
    if( outFile == NULL )
    {
        outFile = stdout;
        if( outFile == NULL )
        {
            fprintf(stderr, "ERROR:  Cannot open stdout for writing\n");
            return(1);
        }
    }


    /*
     * If they want to make an HTML document with style attributes, cool.
     * Otherwise, just replace newlines with <br/> statements as usual.
     */
    if( htmlCreate == 1 )
    {
        html_begin( outFile );
        bloggie( inFile, outFile );
        html_end( outFile );
    }
    else
        bloggie( inFile, outFile );

    /*  Null file pointers */
    inFile = NULL;
    outFile = NULL;

    /*  Free string pointers */
    free( lang );
    free( bgColor );
    free( fgColor );
    free( pageTitle );
    free( fontFamily );
    free( cssFile );
    free( jsFile );

    /*  Null string pointers */
    pageTitle = NULL;
    fontFamily = NULL;
    bgColor = NULL;
    fgColor = NULL;
    cssFile = NULL;
    jsFile = NULL;
    lang = NULL;

    return(0);
}
